@@ -29,6 +29,9 @@
 #include <dbus/dbus-glib-bindings.h>
 #include <dbus/dbus-glib-lowlevel.h>
 #include <dbus/dbus.h>
+#include <gdk/gdk.h>
+#include <gio/gdesktopappinfo.h>
+#include <clutter/x11/clutter-x11.h>
 
 #include "mpl-panel-client.h"
 #include "mpl-panel-common.h"
@@ -89,6 +92,7 @@ struct _MplPanelClientPrivate
   guint            width;
   guint            max_height;
   guint            requested_height;
+  guint            real_height;
 
   gboolean         constructed     : 1; /*poor man's constructor return value*/
   gboolean         toolbar_service : 1;
@@ -245,6 +249,8 @@ mnb_panel_dbus_init_panel (MplPanelClient  *self,
                  priv->requested_height, height);
     }
 
+  priv->real_height = real_height;
+
   *alloc_width  = width;
   *alloc_height = real_height;
 
@@ -264,11 +270,18 @@ mnb_panel_dbus_set_size (MplPanelClient  *self,
 {
   MplPanelClientPrivate *priv = self->priv;
   guint real_height = height;
+  guint old_width = priv->width;
+  guint old_height = priv->real_height;
 
-  g_debug ("%s called: width %d, height %d", __FUNCTION__, width, height);
+  if (height > 0)
+    priv->max_height = height;
+
+  if (width > 0)
+    priv->width = width;
+
+  g_debug ("%s called: width %d (%d), height %d (%d)",
+           __FUNCTION__, width, priv->width, height, priv->max_height);
 
-  priv->max_height = height;
-  priv->width      = width;
 
   if (priv->requested_height > 0 && priv->requested_height < height)
     real_height = priv->requested_height;
@@ -279,7 +292,10 @@ mnb_panel_dbus_set_size (MplPanelClient  *self,
                  priv->requested_height, height);
     }
 
-  g_signal_emit (self, signals[SET_SIZE], 0, width, real_height);
+  priv->real_height = real_height;
+
+  if (old_width != priv->width || old_height != real_height)
+    g_signal_emit (self, signals[SET_SIZE], 0, priv->width, real_height);
 
   return TRUE;
 }
@@ -796,157 +812,224 @@ void
 mpl_panel_client_request_button_style (MplPanelClient *panel,
                                        const gchar    *style)
 {
-  g_signal_emit (panel, signals[REQUEST_BUTTON_STYLE], 0, style);
+  MplPanelClientPrivate *priv = MPL_PANEL_CLIENT (panel)->priv;
+
+  if (!g_str_equal (style, priv->button_style))
+    {
+      g_free (priv->button_style);
+      priv->button_style = g_strdup (style);
+
+      g_signal_emit (panel, signals[REQUEST_BUTTON_STYLE], 0, style);
+    }
 }
 
 void
 mpl_panel_client_request_tooltip (MplPanelClient *panel,
                                   const gchar    *tooltip)
 {
+  MplPanelClientPrivate *priv = MPL_PANEL_CLIENT (panel)->priv;
+
+  g_free (priv->tooltip);
+  priv->tooltip = g_strdup (tooltip);
+
   g_signal_emit (panel, signals[REQUEST_TOOLTIP], 0, tooltip);
 }
 
 #include "mnb-toolbar-dbus-bindings.h"
 
-gboolean
-mpl_panel_client_launch_application (MplPanelClient *panel,
-                                     const gchar    *path,
-                                     gint            workspace,
-                                     gboolean        no_chooser)
+/*
+ * Helper function to launch application from GAppInfo, with all the
+ * required SN housekeeping.
+ */
+static gboolean
+mpl_panel_client_launch_application_from_info (GAppInfo *app, GList *files)
 {
-  MplPanelClientPrivate *priv = panel->priv;
-  GError                *error = NULL;
+  GAppLaunchContext    *ctx;
+  GdkAppLaunchContext  *gctx;
+  GError               *error = NULL;
+  gboolean              retval = TRUE;
+  guint32               timestamp;
 
-  if (!priv->toolbar_proxy)
-    {
-      if (!priv->toolbar_service)
-        g_warning ("Launching API is not available because toolbar services "
-                   "were not enabled at panel construction !!!");
-      else
-        g_warning ("Launching API is not available.");
+  gctx = gdk_app_launch_context_new ();
+  ctx  = G_APP_LAUNCH_CONTEXT (gctx);
 
-      return FALSE;
-    }
+  timestamp = clutter_x11_get_current_event_time ();
 
-  if (!org_moblin_UX_Shell_Toolbar_launch_application (priv->toolbar_proxy,
-                                                       path,
-                                                       workspace,
-                                                       no_chooser,
-                                                       &error))
+  gdk_app_launch_context_set_timestamp (gctx, timestamp);
+
+  retval = g_app_info_launch (app, files, ctx, &error);
+
+  if (!retval)
     {
       if (error)
         {
-          g_warning ("Could not launch application %s: %s",
-                     path, error->message);
+          g_warning ("Failed to launch %s (%s)",
+#if GLIB_CHECK_VERSION(2,20,0)
+                     g_app_info_get_commandline (app),
+#else
+                     g_app_info_get_name (app),
+#endif
+                     error->message);
+
           g_error_free (error);
         }
       else
-        g_warning ("Could not launch application %s", path);
+        {
+          g_warning ("Failed to launch %s",
+#if GLIB_CHECK_VERSION(2,20,0)
+                     g_app_info_get_commandline (app)
+#else
+                     g_app_info_get_name (app)
+#endif
+                     );
+        }
+    }
 
+  g_object_unref (ctx);
+
+  return retval;
+}
+
+gboolean
+mpl_panel_client_launch_application (MplPanelClient *panel, const gchar *path)
+{
+  GAppInfo *app;
+  GError   *error = NULL;
+  gboolean  retval;
+  GKeyFile *key_file;
+  gchar    *cmd;
+
+  g_return_val_if_fail (path, FALSE);
+
+#if 1
+  /*
+   * Startup notification only works with the g_app_launch API when we both
+   * supply a GDK launch context *and* create the GAppInfo from a desktop file
+   * that has the StartupNotify field set to true.
+   *
+   * To work around the limitations, we fake a desktop file via the key-file
+   * API (The alternative is provide a custom GAppInfo impelentation, but the
+   * GAppInfo design makes that hard: g_app_info_create_from_commandline () is
+   * hardcoded to use GDesktopAppInfo, so for any internal glib paths that
+   * pass through it, we are pretty much screwed anyway. It seemed like this
+   * 7LOC hack made more sense than the 800LOC, no less hacky, alternative.
+   */
+  cmd = g_strdup_printf ("%s %%u", path);
+  key_file = g_key_file_new ();
+
+  g_key_file_set_string  (key_file, G_KEY_FILE_DESKTOP_GROUP,
+                          G_KEY_FILE_DESKTOP_KEY_TYPE,
+                          G_KEY_FILE_DESKTOP_TYPE_APPLICATION);
+  g_key_file_set_string  (key_file, G_KEY_FILE_DESKTOP_GROUP,
+                          G_KEY_FILE_DESKTOP_KEY_EXEC, cmd);
+  g_key_file_set_boolean (key_file, G_KEY_FILE_DESKTOP_GROUP,
+                          G_KEY_FILE_DESKTOP_KEY_STARTUP_NOTIFY, TRUE);
+
+  app = (GAppInfo*)g_desktop_app_info_new_from_keyfile (key_file);
+
+  g_key_file_free (key_file);
+  g_free (cmd);
+#else
+  app = g_app_info_create_from_commandline (path, NULL,
+                                            G_APP_INFO_CREATE_SUPPORTS_URIS,
+                                            &error);
+#endif
+
+  if (error)
+    {
+      g_warning ("Failed to create GAppInfo from commnad line %s (%s)",
+                 path, error->message);
+
+      g_error_free (error);
       return FALSE;
     }
 
-  return TRUE;
+  retval = mpl_panel_client_launch_application_from_info (app, NULL);
+
+  g_object_unref (app);
+
+  return retval;
 }
 
 gboolean
 mpl_panel_client_launch_application_from_desktop_file (MplPanelClient *panel,
                                                        const gchar    *desktop,
-                                                       GList          *files,
-                                                       gint            wspace,
-                                                       gboolean      no_chooser)
+                                                       GList          *files)
 {
-  MplPanelClientPrivate *priv = panel->priv;
-  GError                *error = NULL;
-  GList                 *l = files;
-  gchar                 *arguments = NULL;
+  GAppInfo *app;
+  gboolean  retval;
 
-  if (!priv->toolbar_proxy)
+  g_return_val_if_fail (desktop, FALSE);
+
+  app = G_APP_INFO (g_desktop_app_info_new_from_filename (desktop));
+
+  if (!app)
     {
-      if (!priv->toolbar_service)
-        g_warning ("Launching API is not available because toolbar services "
-                   "were not enabled at panel construction !!!");
-      else
-        g_warning ("Launching API is not available.");
-
+      g_warning ("Failed to create GAppInfo for file %s", desktop);
       return FALSE;
     }
 
-  while (l)
-    {
-      if (l->data)
-        {
-          gchar *a = g_strconcat (arguments, " ", l->data, NULL);
-          g_free (arguments);
-          arguments = a;
-        }
-      l = l->next;
-    }
+  retval = mpl_panel_client_launch_application_from_info (app, files);
 
-  if (!org_moblin_UX_Shell_Toolbar_launch_application_by_desktop_file (
-                                                        priv->toolbar_proxy,
-                                                        desktop,
-                                                        arguments,
-                                                        wspace,
-                                                        no_chooser,
-                                                        &error))
-    {
-      if (error)
-        {
-          g_warning ("Could not launch application from desktop file %s: %s",
-                     desktop, error->message);
-          g_error_free (error);
-        }
-      else
-        g_warning ("Could not launch application from desktop file %s",
-                   desktop);
+  g_object_unref (app);
 
-      return FALSE;
-    }
-
-  return TRUE;
+  return retval;
 }
 
 gboolean
 mpl_panel_client_launch_default_application_for_uri (MplPanelClient *panel,
-                                                     const gchar    *uri,
-                                                     gint            workspace,
-                                                     gboolean        no_chooser)
+                                                     const gchar    *uri)
 {
-  MplPanelClientPrivate *priv = panel->priv;
-  GError                *error = NULL;
+  GAppLaunchContext   *ctx;
+  GdkAppLaunchContext *gctx;
+  GAppInfo            *app;
+  GError              *error = NULL;
+  gboolean             retval = TRUE;
+  gchar               *uri_scheme;
+  guint32              timestamp;
 
-  if (!priv->toolbar_proxy)
+  uri_scheme = g_uri_parse_scheme (uri);
+
+  /* For local files we want the local file handler not the scheme handler */
+  if (g_str_equal (uri_scheme, "file"))
     {
-      if (!priv->toolbar_service)
-        g_warning ("Launching API is not available because toolbar services "
-                   "were not enabled at panel construction !!!");
-      else
-        g_warning ("Launching API is not available.");
-
-      return FALSE;
+      GFile *file;
+      file = g_file_new_for_uri (uri);
+      app = g_file_query_default_handler (file, NULL, NULL);
+      g_object_unref (file);
+    }
+  else
+    {
+      app = g_app_info_get_default_for_uri_scheme (uri_scheme);
     }
 
-  if (!org_moblin_UX_Shell_Toolbar_launch_default_application_for_uri (
-                                                  priv->toolbar_proxy,
-                                                  uri,
-                                                  workspace,
-                                                  no_chooser,
-                                                  &error))
+  g_free (uri_scheme);
+
+  gctx = gdk_app_launch_context_new ();
+  ctx  = G_APP_LAUNCH_CONTEXT (gctx);
+
+  timestamp = clutter_x11_get_current_event_time ();
+
+  gdk_app_launch_context_set_timestamp (gctx, timestamp);
+
+  retval = g_app_info_launch_default_for_uri (uri, ctx, &error);
+
+  if (!retval)
     {
       if (error)
         {
-          g_warning ("Could not launch default application for uri %s: %s",
+          g_warning ("Failed to launch default app for %s (%s)",
                      uri, error->message);
+
           g_error_free (error);
         }
       else
-        g_warning ("Could not launch default application for %s", uri);
-
-      return FALSE;
+        g_warning ("Failed to launch default app for %s", uri);
     }
 
-  return TRUE;
+  g_object_unref (ctx);
+
+  return retval;
 }
 
 void
