@@ -35,6 +35,7 @@
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <moblin-panel/mpl-panel-common.h>
+#include <display.h>
 
 G_DEFINE_TYPE (MnbPanel, mnb_panel, MNB_TYPE_DROP_DOWN)
 
@@ -144,14 +145,35 @@ mnb_panel_set_property (GObject *object, guint property_id,
     }
 }
 
+void
+mnb_panel_focus (MnbPanel *panel)
+{
+  MetaDisplay *display;
+  MutterPlugin *plugin = NULL;
+
+  g_object_get (G_OBJECT (panel), "mutter-plugin", &plugin, NULL);
+  if (!plugin)
+        return;
+
+  display = meta_screen_get_display (mutter_plugin_get_screen (plugin));
+
+  gdk_error_trap_push ();
+
+  XRaiseWindow (meta_display_get_xdisplay (display), panel->priv->xid);
+
+  XSetInputFocus (meta_display_get_xdisplay (display), panel->priv->child_xid,
+                  RevertToPointerRoot, CurrentTime);
+
+  gdk_flush ();
+  gdk_error_trap_pop ();
+}
+
 /*
  * Callbacks for the request signals exposed by the panels.
  */
 static void
 mnb_panel_request_focus_cb (DBusGProxy *proxy, MnbPanel *panel)
 {
-  MnbPanelPrivate *priv;
-
   if (!CLUTTER_ACTOR_IS_MAPPED (panel))
     {
       g_warning ("Panel %s requested focus while not visible !!!",
@@ -159,15 +181,7 @@ mnb_panel_request_focus_cb (DBusGProxy *proxy, MnbPanel *panel)
       return;
     }
 
-  priv = panel->priv;
-
-  XRaiseWindow (GDK_DISPLAY(),
-                priv->xid);
-
-  XSetInputFocus (GDK_DISPLAY(),
-                  priv->child_xid,
-                  RevertToPointerRoot,
-                  CurrentTime);
+  mnb_panel_focus (panel);
 }
 
 static void
@@ -215,18 +229,29 @@ mnb_panel_set_size_cb (DBusGProxy *proxy,
                        MnbPanel   *panel)
 {
   MnbPanelPrivate *priv = panel->priv;
+  GtkWidget       *socket;
+  GdkWindow       *socket_win;
 
   if (!priv->window)
     return;
 
-  g_debug (G_STRLOC " Got size change on Panel %s",
-           mnb_panel_get_name (panel));
+  priv->width  = width;
+  priv->height = height;
 
   /*
-   * Resize our top-level window to match.
+   * Resize our top-level window to match. If the window is hidden, move it
+   * so that we can be sure it is entirely off screen.
    */
+  if (!CLUTTER_ACTOR_IS_VISIBLE (panel))
+    gtk_window_move (GTK_WINDOW (priv->window),
+                     TOOLBAR_X_PADDING, -(height + 100));
+
   gtk_window_resize (GTK_WINDOW (priv->window), width, height);
 
+  if ((socket = gtk_bin_get_child (GTK_BIN (priv->window))) &&
+      (socket_win = gtk_socket_get_plug_window (GTK_SOCKET (socket))))
+    gdk_window_resize (socket_win, width, height);
+
   g_signal_emit (panel, signals[SET_SIZE], 0, width, height);
 }
 
@@ -462,15 +487,10 @@ mnb_panel_show_completed (MnbDropDown *self)
 
   clutter_actor_get_position (CLUTTER_ACTOR (self), &x, &y);
 
-  gtk_window_move (GTK_WINDOW (priv->window), (gint)x, (gint)y);
+  gtk_window_move (GTK_WINDOW (priv->window),
+                   (gint)x + TOOLBAR_X_PADDING, (gint)y);
 
-  XRaiseWindow (GDK_DISPLAY(),
-                priv->xid);
-
-  XSetInputFocus (GDK_DISPLAY(),
-                  priv->child_xid,
-                  RevertToPointerRoot,
-                  CurrentTime);
+  mnb_panel_focus (MNB_PANEL (self));
 
   org_moblin_UX_Shell_Panel_show_end_async (priv->proxy,
                                             mnb_panel_dbus_dumb_reply_cb, NULL);
@@ -494,7 +514,8 @@ mnb_panel_hide_begin (MnbDropDown *self)
    * We do not hide the window here, only move it off screen; this significantly
    * improves the Toolbar button response time.
    */
-  gtk_window_move (GTK_WINDOW (window), 0, -(priv->height + 100));
+  gtk_window_move (GTK_WINDOW (window), TOOLBAR_X_PADDING,
+                   -(priv->height + 100));
 
   org_moblin_UX_Shell_Panel_hide_begin_async (priv->proxy,
                                               mnb_panel_dbus_dumb_reply_cb, NULL);
@@ -542,11 +563,7 @@ mnb_panel_socket_size_allocate_cb (GtkWidget     *widget,
 {
   GtkSocket *socket = GTK_SOCKET (widget);
 
-  if (!socket->is_mapped)
-    {
-      g_debug ("socket deallocated");
-    }
-  else
+  if (socket->is_mapped)
     {
       MnbPanelPrivate *priv = MNB_PANEL (data)->priv;
 
@@ -558,16 +575,11 @@ mnb_panel_socket_size_allocate_cb (GtkWidget     *widget,
         {
           gfloat x = 0, y = 0;
 
-          g_debug ("socket allocated %d,%d;%dx%d",
-                   allocation->x,
-                   allocation->y,
-                   allocation->width,
-                   allocation->height);
-
           clutter_actor_get_position (CLUTTER_ACTOR (data), &x, &y);
 
-          gtk_window_move (GTK_WINDOW (priv->window), (gint)x, (gint)y +
-                           MNB_DROP_DOWN_TOP_PADDING);
+          gtk_window_move (GTK_WINDOW (priv->window),
+                           (gint)x + TOOLBAR_X_PADDING,
+                           (gint)y + MNB_DROP_DOWN_TOP_PADDING);
         }
     }
 }
@@ -663,7 +675,8 @@ mnb_panel_init_panel_reply_cb (DBusGProxy *proxy,
   gtk_window_resize (GTK_WINDOW (window), window_width, window_height);
 
   gtk_window_move (GTK_WINDOW (window),
-                   0, TOOLBAR_HEIGHT + MNB_DROP_DOWN_TOP_PADDING);
+                   TOOLBAR_X_PADDING,
+                   TOOLBAR_HEIGHT + MNB_DROP_DOWN_TOP_PADDING);
 
   gtk_container_add (GTK_CONTAINER (window), socket);
   gtk_widget_realize (socket);
@@ -723,9 +736,6 @@ mnb_panel_proxy_owner_changed_cb (DBusGProxy *proxy,
                                   const char *new,
                                   gpointer    data)
 {
-  g_debug ("Proxy %s owner changed (old %s, new %s).",
-           name, old, new);
-
   mnb_panel_init_owner (MNB_PANEL (data));
 }
 
@@ -743,9 +753,6 @@ mnb_panel_setup_proxy (MnbPanel *panel)
       return FALSE;
     }
 
-  g_debug ("Creating proxy for %s, %p",
-           priv->dbus_name, panel);
-
   dbus_path = g_strconcat ("/", priv->dbus_name, NULL);
 
   p = dbus_path;
@@ -963,10 +970,7 @@ mnb_panel_show_mutter_window (MnbPanel *panel, MutterWindow *mcw)
     }
 
   if (mcw == priv->mcw)
-    {
-      g_debug ("Window already handled.");
-      return;
-    }
+    return;
 
   texture = mutter_window_get_texture (mcw);
 
@@ -1013,23 +1017,56 @@ mnb_panel_set_size (MnbPanel *panel, guint width, guint height)
 {
   MnbPanelPrivate *priv = panel->priv;
   gfloat x, y, w, h;
+  gint   wi, hi, hfi;
+  gboolean h_change = FALSE, w_change = FALSE;
 
   /*
    * Exit if no change
    */
   clutter_actor_get_size (CLUTTER_ACTOR (panel), &w, &h);
 
-  if (((guint)w == width) && ((guint)h == height))
+  if ((guint)w != width)
+    w_change = TRUE;
+
+  if ((guint)h != height)
+    h_change = TRUE;
+
+  if (!w_change && !h_change)
     return;
 
   mnb_drop_down_get_footer_geometry (MNB_DROP_DOWN (panel), &x, &y, &w, &h);
 
+  /*
+   * Hack: the footer geometry API for some reason returns 0 the first time
+   * we try to show the panel. Probably the panel is not yet fully setup. As
+   * a quick fix, fallback to the default size here, if that happens.
+   */
+  hfi = (gint) h;
+
+  if (!hfi)
+    hfi = TOOLBAR_HEIGHT/2;
+
   /*
    * FIXME -- the border and shaddow should not be hardcoded here.
    */
-  org_moblin_UX_Shell_Panel_set_size_async (priv->proxy,
-                                            width - 2 * 4,
-                                            height - (guint) h - 4 - 37,
+  wi = width - 2 * 4;
+  hi = height - hfi - MNB_DROP_DOWN_TOP_PADDING - MNB_DROP_DOWN_SHADOW_HEIGHT;
+
+  if (wi < 0)
+    wi = 0;
+
+  if (hi < 0)
+    hi = 0;
+
+  org_moblin_UX_Shell_Panel_set_size_async (priv->proxy, wi, hi,
                                             mnb_panel_dbus_dumb_reply_cb, NULL);
 }
 
+MutterWindow *
+mnb_panel_get_mutter_window (MnbPanel *panel)
+{
+  MnbPanelPrivate *priv = panel->priv;
+
+  return priv->mcw;
+}
+
